/*
 * File:   log.hh
 * Author: Balint Morvai
 *
 * Modified on Jan 2013
 *
 * Usage: 	The "log" class builds up on the "clog" file and stream 
 * 			classes. It provides numerous operators and vars to control
 * 			the output that is directed to the "clog" classes. Optionally
 * 			it also uses the "boost signals2" log class.
 * 			For further settings also see "log_macros.hh".
 * 
 * MODIFIED BSD LICENSE:
 * 
 * Copyright (c) 2013, Balint Morvai
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the "Logger" nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef LOG_HH
#define	LOG_HH

#include <boost/signals2.hpp>
#include "clog.hh"
#include "log_macros.hh"

#ifndef MPIHELPER
#define MPIHELPER
// DUMMY class used below to get processor id!
// REPLACE THIS CLASS WITH YOUR OWN MPI HELPER CLASS THAT
// WILL RETURN CORRECT RANK.
// BE SURE TO DO IT ALSO IN FILE "clog.hh"!
class MPIhelper
{
	MPIhelper(){};
	~MPIhelper(){};
	
	int rank()
	{
		return 1;
	};
};
#endif

class Log
{
	private:
	boost::signals2::signal<void(std::string const&)> boostlog;
	boost::signals2::connection boost_cf; // connection to file stream
	boost::signals2::connection boost_cs; // connection to string stream
	MPIhelper		 	*helper;// MPI helper class that must contain a "rank()" method returning an integer!
	CLogfile 			clogf;	// writes to specific file for each process
	CLogstream 			clogs;	// writes to common stream for each process
	logout 				out; 	// sets where the log output goes
	int					pid;	// sets on which processor to output
	
	public:
	bool show_time;	// tells whether to show time info
	bool show_pid;	// tells whether to show pid info in stream
	bool info_sticky; // tells whether to keep showing info once an out with "i"
					 // has been set or to drop the "i" after the first output.

	Log( std::ofstream *logout_f, std::ostream *logout_s )
			: 	clogf(CLogfile(logout_f)), 
				clogs(CLogstream(logout_s)), 
				out(OFS),
				pid(-1),
				show_time(true), 
				show_pid(true),
				info_sticky(true)
	{
		// connect signals2 slots to file and stream logs - since non static class members can not be
		// passed by reference, we us a complicated trick: boost::bind passes to param1 all further params
		// and returns a func handle to a func that executes the call. E.g: bind(sum,1,2) returns a func 
		// calling sum(1,2). Here more complex: 1st param is pointer, thus 2nd gets put inside and third
		// is a placeholder that gets passed 1st argument passed to func.
		boost_cf = boostlog.connect(boost::bind(&CLogfile::slotf, clogf, _1));
		boost_cs = boostlog.connect(boost::bind(&CLogstream::slots, clogs, _1));
	};

	Log& operator[] (logout out_) 
	{
		set_logout(out_);
		
		return *this;
	};

	Log& operator() (int pid_ = -1) 
	{
		pid = pid_;
		
		return *this;
	};

	// Thread safe boost version. Use this when more processes are 
	// sharing the same external streams.
	// NOTE: the boost version of std modifier support will only work for
	//		 the few std modifiers producing output like "std::endl". The others
	//		 wont have an effect.
	Log& operator<(std::ostream& (*pf)(std::ostream&))
	{
		std::stringstream msg_new;
		logout subout = OFS;

		// disconnect file or stream outs from boost when necessary
		// let check_pid check whether output on this processor and
		// let is_equal check whether output to this stream
		if(!check_pid(OF) || is_equal(out, OS))
		{
			boost_cf.disconnect();
			subout = OS;
		}
		if(!check_pid(OS) || is_equal(out, OF))
		{
			boost_cs.disconnect();
			subout = OF;
		}
		// send message	
		// dont use get_prefix since we dont want info with modifiers
		msg_new << pf;
		boostlog( msg_new.str() );
		// reconnect boost
		if(subout == OS)
			boost_cf = boostlog.connect(boost::bind(&CLogfile::slotf, clogf, _1));
		if(subout == OF)
			boost_cs = boostlog.connect(boost::bind(&CLogstream::slots, clogs, _1));

		if(!info_sticky)
			out--; // strip "i": reduce "out" case to the one equal but without "i"
			
		return *this;
	};
	
	template < typename T >
	Log& operator<( T const & t ) 
	{ 
		std::stringstream msg_new;
		logout subout = OFS;

		// disconnect file or stream outs from boost when necessary
		// let check_pid check whether output on this processor and
		// let is_equal check whether output to this stream
		if(!check_pid(OF) || is_equal(out, OS))
		{
			boost_cf.disconnect();
			subout = OS;
		}
		if(!check_pid(OS) || is_equal(out, OF))
		{
			boost_cs.disconnect();
			subout = OF;
		}
		// send message	
		// let get_prefix determin whether to show info and what to show
		msg_new << get_prefix(subout) << t;
		boostlog( msg_new.str() );
		// reconnect boost
		if(subout == OS)
			boost_cf = boostlog.connect(boost::bind(&CLogfile::slotf, clogf, _1));
		if(subout == OF)
			boost_cs = boostlog.connect(boost::bind(&CLogstream::slots, clogs, _1));

		if(!info_sticky)
			out--; // strip "i": reduce "out" case to the one equal but without "i"
			
		return *this;
	};

	Log& operator<<(std::ostream& (*pf)(std::ostream&))
	{
		// dont use get_prefix since we dont want info with modifiers
		// let check_pid check whether output on this processor and
		// let is_equal check whether output to this stream
		if(check_pid(OF) && !is_equal(out, OS))
			clogf << pf;
		if(check_pid(OS) && !is_equal(out, OF))
			clogs << pf;
		if(!info_sticky)
			out--; // strip "i": reduce "out" case to the one equal but without "i"

		return *this;
	};

	template < typename T >
	Log& operator<<( T const & t )
	{
		// let get_prefix determin whether to show info and what to show
		// let check_pid check whether output on this processor and
		// let is_equal check whether output to this stream
		if(check_pid(OF) && !is_equal(out, OS))
			clogf << get_prefix(OF) << t;
		if(check_pid(OS) && !is_equal(out, OF))
			clogs << get_prefix(OS) << t;
		if(!info_sticky)
			out--; // strip "i": reduce "out" case to the one equal but without "i"
		
		return *this;
	};

	CLogfile& get_clogf_inst()
	{
		return clogf;
	};
	
	CLogstream& get_clogs_inst()
	{
		return clogs;
	};

	void set_logout(logout out_)
	{
		out = out_;
	};

	// use the parameter to tell if the check is for file or stream out:
	// the check is NOT done, thus always TRUE, for certain combinations
	// of subout and this->out. (depending on "p" in out and on subout)
	// NOTE: if subout contains "FS" the check is only omitted if
	//		 for BOTH not required by setting of "p" in out. Thus mostly
	//		 you will want to use "OS" and "OF" macros only as param.
	bool check_pid(logout subout = OS)
	{
		bool pass = false;
		bool const check = (helper->rank() == pid);
		
		if(pid < 0) pass = true;
		else
		{
			switch(out)
			{	// combine normal & i cases, get_prefix checks whether to show info
				// NOTE: is_equal(subout, X) means if subout = X ignoring "i" / "p" in X
				case OFS: case OFSi: case OFpSpi : case OFpSp:
					pass = check;
					break;
				case OFpS: case OFpSi:
					if(!is_equal(subout, OS))
						pass = check;
					else
						pass = true;
					break;
				case OFSp: case OFSpi:
					if(!is_equal(subout, OF))
						pass = check;
					else
						pass = true;
					break;    
				case OF: case OFi:
					if(is_equal(subout, OF))
						pass = check;
					break;
				case OS: case OSi: 
					if(is_equal(subout, OS))
						pass = check;
					break;
				default: break; // just to satisfy compiler (no warning)
			}
		}

		return pass;
	};

	// use the parameter to tell if the prefix is for file or stream out:
	// the rank info is NOT included for file only subouts OF & OFi, but
	// included for all other subouts.
	std::string get_prefix(logout subout = OS)
	{
		std::stringstream msg_new;
		
		//msg_new << "get_prefix; out is " << out << " and ...";
		if(is_i(out)) // if out has "i"
		{
			//msg_new << "it passes! ";
			if(show_time)
				msg_new << get_time() 	<< LOG_SEPARATOR;
			if(show_pid && !is_equal(subout, OF)) // if subout = OF ignoring "i" / "p"
				msg_new << get_rank() 	<< LOG_SEPARATOR;
		}
			
		return msg_new.str();
	};

	std::string get_rank()
	{
		// rank vars
		std::stringstream rank_stream;
		
		rank_stream << LOG_RANK_STRING << helper->rank();
		
		return rank_stream.str();
	};
	
	std::string get_time()
	{
		// time vars
		time_t 				time_raw;
		struct tm * 		time_info;
		char				time_buff[50];
		std::stringstream 	time_stream;

		time( &time_raw );
		time_info = localtime ( &time_raw );
		strftime( time_buff, 50, LOG_TIME_FORMAT, time_info );
		time_stream << time_buff;
		
		return time_stream.str();
	};

	void init(	Dune::MPIHelper *helper_, 
				const std::string dir = "", 
				const std::string filename = "", 
				bool show_time_ = true, 
				bool show_pid_ = true,
				bool info_sticky_ = false)
	{
		std::stringstream logpath;
		helper = helper_;
		show_time = show_time_;
		show_pid = show_pid_;
		info_sticky = info_sticky_;
		
		// if a dir was given - corresponds to "else" case of filename
		if(dir.compare("")) logpath << dir << "/";
		// if a filename was given
		if(!filename.compare(""))
		{	// construct filename of process-specific logfiles
			logpath << LOG_FILE_NAME << helper->rank() << LOG_FILE_EXT;
		}
		else logpath << filename;

		// init file and stream clogs (this is not done through their constructors)
		clogf.init(helper, logpath.str());
		clogs.init(helper);
	};
	
	~Log() 
	{};
	
	
};

#endif // LOG_HH

